from typing import Any, get_origin, get_type_hints


def clean_data(model: type[object], data: dict[str, Any]) -> dict[str, Any]:
    hints = get_type_hints(model)
    return {key: _coerce(hints[key], value) for key, value in data.items()}


def _coerce(param_type, value: Any) -> Any:
    if (
        get_origin(param_type) is not list
        and isinstance(value, list)
        and len(value) == 1
    ):
        value = value[0]

    return value



from __future__ import annotations

from dataclasses import is_dataclass
from inspect import isclass, iscoroutine
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    get_args,
    get_type_hints,
)

from sanic import Request
from sanic.app import Sanic
from sanic.exceptions import ServerError

from sanic_ext.exceptions import InitError
from sanic_ext.utils.typing import (
    is_attrs,
    is_msgspec,
    is_optional,
    is_pydantic,
)


if TYPE_CHECKING:
    from .registry import ConstantRegistry, InjectionRegistry


class Constructor:
    EXEMPT_ANNOTATIONS = (Request,)

    def __init__(
        self, func: Callable[..., Any], request_arg: Optional[str] = None
    ):
        self.func = func
        self.injections: dict[str, tuple[type, Constructor]] = {}
        self.constants: dict[str, Any] = {}
        self.pass_kwargs: bool = False
        self.request_arg = request_arg

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}:{self.func.__name__}>"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(func={self.func.__name__})>"

    async def __call__(self, request, **kwargs):
        try:
            args = await gather_args(self.injections, request, **kwargs)
            args.update(self.constants)
            if self.pass_kwargs:
                args.update(kwargs)

            if self.request_arg:
                args[self.request_arg] = request

            retval = self.func(**args)

            if iscoroutine(retval):
                retval = await retval
            return retval
        except TypeError as e:
            raise ServerError(
                "Failure to inject dependencies. Make sure that all "
                f"dependencies for '{self.func.__name__}' have been "
                "registered."
            ) from e

    def prepare(
        self,
        app: Sanic,
        injection_registry: InjectionRegistry,
        constant_registry: ConstantRegistry,
        allowed_types: set[type[object]],
    ) -> None:
        hints = self._get_hints()
        hints.pop("return", None)
        missing = []
        for param, annotation in hints.items():
            if param in constant_registry:
                self.constants[param] = getattr(app.config, param.upper())
            if annotation in allowed_types:
                self.pass_kwargs = True
            if is_optional(annotation):
                annotation = get_args(annotation)[0]
            if not isclass(annotation):
                missing.append((param, annotation))
                continue
            if issubclass(annotation, Request):
                self.request_arg = param
            if (
                annotation not in self.EXEMPT_ANNOTATIONS
                and not issubclass(annotation, self.EXEMPT_ANNOTATIONS)
                and annotation not in allowed_types
            ):
                dependency = injection_registry.get(annotation)
                if not dependency:
                    missing.append((param, annotation))
                    continue
                self.injections[param] = (annotation, dependency)

        if missing:
            dependencies = "\n".join(
                [f"  - {param}: {annotation}" for param, annotation in missing]
            )
            raise InitError(
                "Unable to resolve dependencies for "
                f"'{self.func.__name__}'. Could not find the following "
                f"dependencies:\n{dependencies}.\nMake sure the dependencies "
                "are declared using ext.injection. See "
                "https://sanicframework.org/en/plugins/sanic-ext/injection."
                "html#injecting-services for more details."
            )

        checked: set[type[object]] = set()
        current: set[type[object]] = set()
        self.check_circular(checked, current)

    def check_circular(
        self,
        checked: set[type[object]],
        current: set[type[object]],
    ) -> None:
        dependencies = set(self.injections.values())
        for dependency, constructor in dependencies:
            self._visit(dependency, constructor, checked, current)

    def _visit(
        self,
        dependency: type[object],
        constructor: Constructor,
        checked: set[type[object]],
        current: set[type[object]],
    ):
        if dependency in checked:
            return
        elif dependency in current:
            raise InitError(
                "Circular dependency injection detected on "
                f"'{self.func.__name__}'. Check dependencies of "
                f"'{constructor.func.__name__}' which may contain "
                f"circular dependency chain with {dependency}."
            )

        current.add(dependency)
        constructor.check_circular(checked, current)
        current.remove(dependency)
        checked.add(dependency)

    def _get_hints(self):
        if (
            not isclass(self.func)
            or is_dataclass(self.func)
            or is_attrs(self.func)
            or is_pydantic(self.func)
            or is_msgspec(self.func)
        ):
            return get_type_hints(self.func)
        elif isclass(self.func):
            return get_type_hints(self.func.__init__)
        raise InitError(f"Cannot get type hints for {self.func}")


async def gather_args(injections, request, **kwargs) -> dict[str, Any]:
    return {
        name: await do_cast(_type, constructor, request, **kwargs)
        for name, (_type, constructor) in injections.items()
    }


async def do_cast(_type, constructor, request, **kwargs):
    cast = constructor if constructor else _type
    args = [request] if constructor else []

    retval = cast(*args, **kwargs)
    if iscoroutine(retval):
        retval = await retval
    return retval



from dataclasses import dataclass, field
from typing import Annotated
from uuid import UUID

import attrs
import pytest

from msgspec import Meta, Struct
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass as pydataclass

from sanic_ext import openapi

from .utils import get_spec


@dataclass
class FooDataclass:
    links: list[UUID]
    priority: int = field(
        metadata={"openapi": {"exclusiveMinimum": 1, "exclusiveMaximum": 10}}
    )
    ident: str = field(
        default="XXXX", metadata={"openapi": {"example": "ABC123"}}
    )


@attrs.define
class FooAttrs:
    links: list[UUID]
    priority: int = attrs.field(
        metadata={"openapi": {"exclusiveMinimum": 1, "exclusiveMaximum": 10}}
    )
    ident: str = attrs.field(
        default="XXXX", metadata={"openapi": {"example": "ABC123"}}
    )


class FooPydanticBaseModel(BaseModel):
    links: list[UUID]
    priority: int = Field(gt=1, lt=10)
    ident: str = Field("XXXX", example="ABC123")


@pydataclass
class FooPydanticDataclass:
    links: list[UUID]
    priority: int = Field(gt=1, lt=10)
    ident: str = Field("XXXX", example="ABC123")


class FooStruct(Struct):
    links: list[UUID]
    priority: Annotated[
        int,
        Meta(
            extra={"openapi": {"exclusiveMinimum": 1, "exclusiveMaximum": 10}}
        ),
    ]
    ident: Annotated[str, Meta(extra={"openapi": {"example": "ABC123"}})] = (
        "XXXX"
    )


models = [
    FooDataclass,
    FooAttrs,
    FooPydanticBaseModel,
    FooPydanticDataclass,
]

models.append(FooStruct)


@pytest.mark.parametrize("Foo", models)
def test_models(app, Foo):
    @app.get("/")
    @openapi.definition(body={"application/json": Foo})
    async def handler(_): ...

    spec = get_spec(app)
    foo_props = spec["paths"]["/"]["get"]["requestBody"]["content"][
        "application/json"
    ]["schema"]["properties"]

    assert foo_props["links"] == {
        "title": "Links",
        "type": "array",
        "items": {"type": "string", "format": "uuid"},
    }
    assert foo_props["ident"] == {
        "title": "Ident",
        "type": "string",
        "default": "XXXX",
        "example": "ABC123",
    }
    assert foo_props["priority"] == {
        "title": "Priority",
        "type": "integer",
        "format": "int32",
        "exclusiveMinimum": 1,
        "exclusiveMaximum": 10,
    }



import logging

from collections import defaultdict
from logging import LogRecord
from logging.handlers import QueueHandler
from multiprocessing import Manager
from queue import Empty, Full
from signal import SIGINT, SIGTERM
from signal import signal as signal_func

from sanic import Sanic
from sanic.log import logger as root_logger
from sanic.log import logger as server_logger
from sanic.logging.setup import setup_logging

from sanic_ext.extensions.logging.extractor import LoggingConfigExtractor


async def prepare_logger(app: Sanic, *_):
    Logger.prepare(app)


async def setup_logger(app: Sanic, *_):
    logger = Logger()
    extractor = LoggingConfigExtractor()
    for logger_name in app.config.LOGGERS:
        extractor.add_logger(logging.getLogger(logger_name))
    app.manager.manage(
        "Logger",
        logger,
        {
            "queue": app.shared_ctx.logger_queue,
            "config": extractor.compile(),
        },
        transient=True,
    )


class SanicQueueHandler(QueueHandler):
    def emit(self, record: LogRecord) -> None:
        try:
            return super().enqueue(record)
        except Full:
            server_logger.warning(
                "Background logger is full. Emitting log in process."
            )
            server_logger.handle(record)


async def setup_server_logging(app: Sanic):
    qhandler = SanicQueueHandler(app.shared_ctx.logger_queue)
    app.ctx._logger_handlers = defaultdict(list)
    app.ctx._qhandler = qhandler

    for logger_name in app.config.LOGGERS:
        logger_instance = logging.getLogger(logger_name)
        for handler in logger_instance.handlers:
            logger_instance.removeHandler(handler)
        logger_instance.addHandler(qhandler)


async def remove_server_logging(app: Sanic):
    for logger, handlers in app.ctx._logger_handlers.items():
        logger.removeHandler(app.ctx._qhandler)
        for handler in handlers:
            logger.addHandler(handler)


class Logger:
    LOGGERS: list[str] = []

    def __init__(self):
        self.run = True
        self.loggers = {
            logger: logging.getLogger(logger) for logger in self.LOGGERS
        }

    def __call__(self, queue, config) -> None:
        signal_func(SIGINT, self.stop)
        signal_func(SIGTERM, self.stop)

        logging.config.dictConfig(config)

        setup_loggers = set(config["loggers"].keys())
        enabled_loggers = set(self.loggers.keys())
        missing = enabled_loggers - setup_loggers
        root_logger.info(
            f"Setup background logging for: {', '.join(setup_loggers)}"
        )
        if missing:
            root_logger.warning(
                f"Logger config not found for: {', '.join(missing)}"
            )
        setup_logging(True, no_color=False, log_extra=True)

        while self.run:
            try:
                record: LogRecord = queue.get(timeout=0.05)
            except Empty:
                continue
            logger = self.loggers.get(record.name)
            logger.handle(record)

    def stop(self, *_):
        if self.run:
            self.run = False

    @classmethod
    def update_cls_loggers(cls, logger_names: list[str]):
        cls.LOGGERS = logger_names

    @classmethod
    def prepare(cls, app: Sanic):
        sync_manager = Manager()
        logger_queue = sync_manager.Queue(
            maxsize=app.config.LOGGING_QUEUE_MAX_SIZE
        )
        app.shared_ctx.logger_queue = logger_queue
        cls.update_cls_loggers(app.config.LOGGERS)

    @classmethod
    def setup(cls, app: Sanic):
        app.main_process_start(prepare_logger)
        app.main_process_ready(setup_logger)
        app.before_server_start(setup_server_logging)
        app.before_server_stop(remove_server_logging)



import logging

from typing import Any, Optional, TypedDict


class LoggerConfig(TypedDict):
    level: str
    propagate: bool
    handlers: list[str]


class HandlerConfig(TypedDict):
    class_: str
    level: str
    stream: Optional[str]
    formatter: Optional[str]


class FormatterConfig(TypedDict):
    class_: str
    format: Optional[str]
    datefmt: Optional[str]


class LoggingConfig(TypedDict):
    version: int
    disable_existing_loggers: bool
    formatters: dict[str, FormatterConfig]
    handlers: dict[str, HandlerConfig]
    loggers: dict[str, LoggerConfig]


class LoggingConfigExtractor:
    def __init__(self):
        self.version = 1
        self.disable_existing_loggers = False
        self.formatters: dict[str, FormatterConfig] = {}
        self.handlers: dict[str, HandlerConfig] = {}
        self.loggers: dict[str, LoggerConfig] = {}

    def add_logger(self, logger: logging.Logger):
        self._extract_logger_config(logger)
        self._extract_handlers(logger)

    def compile(self) -> LoggingConfig:
        output = {
            "version": self.version,
            "disable_existing_loggers": self.disable_existing_loggers,
            "formatters": self.formatters,
            "handlers": self.handlers,
            "loggers": self.loggers,
        }
        return self._clean(output)

    def _extract_logger_config(self, logger: logging.Logger):
        config: LoggerConfig = {
            "level": logging.getLevelName(logger.level),
            "propagate": logger.propagate,
            "handlers": [handler.get_name() for handler in logger.handlers],
        }
        self.loggers[logger.name] = config

    def _extract_handlers(self, logger: logging.Logger):
        for handler in logger.handlers:
            self._extract_handler_config(handler)

    def _extract_handler_config(self, handler: logging.Handler):
        handler_name = handler.get_name()
        if handler_name in self.handlers:
            return
        config: HandlerConfig = {
            "class_": self._full_name(handler),
            "level": logging.getLevelName(handler.level),
            "formatter": (
                self._formatter_name(handler.formatter)
                if handler.formatter
                else None
            ),
            "stream": None,
        }
        # if (stream := getattr(handler, "stream", None)) and (
        #     stream_name := getattr(stream, "name", None)
        # ):
        #     config["stream"] = stream_name
        self.handlers[handler_name] = config
        if handler.formatter:
            self._extract_formatter_config(handler.formatter)

    def _extract_formatter_config(self, formatter: logging.Formatter):
        formatter_name = self._formatter_name(formatter)
        if formatter_name in self.formatters:
            return
        config: FormatterConfig = {
            "class_": self._full_name(formatter),
            "format": formatter._fmt,
            "datefmt": formatter.datefmt,
        }
        self.formatters[formatter_name] = config

    def _clean(self, d: dict[str, Any]) -> dict[str, Any]:
        return {
            k.replace("class_", "class"): self._clean(v)
            if isinstance(v, dict)
            else v
            for k, v in d.items()
        }

    @staticmethod
    def _formatter_name(
        formatter: logging.Formatter, prefix: str = "formatter"
    ):
        return f"{prefix}_{formatter.__class__.__name__}".lower()

    @staticmethod
    def _full_name(obj):
        return f"{obj.__module__}.{obj.__class__.__name__}"



"""
This module provides decorators which append
documentation to OperationStore() and components created in the blueprints.

"""

from collections.abc import Sequence
from functools import wraps
from inspect import isawaitable, isclass
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from sanic import Blueprint
from sanic.exceptions import InvalidUsage, SanicException

from sanic_ext.extensions.openapi import definitions
from sanic_ext.extensions.openapi.builders import (
    OperationStore,
    SpecificationBuilder,
)
from sanic_ext.extensions.openapi.definitions import Component
from sanic_ext.extensions.openapi.types import (
    Array,
    Binary,
    Boolean,
    Byte,
    Date,
    DateTime,
    Double,
    Email,
    Float,
    Integer,
    Long,
    Object,
    Password,
    Schema,
    String,
    Time,
)
from sanic_ext.extras.validation.setup import do_validation, generate_schema
from sanic_ext.utils.extraction import extract_request


__all__ = (
    "definitions",
    "body",
    "component",
    "definition",
    "deprecated",
    "description",
    "document",
    "exclude",
    "no_autodoc",
    "operation",
    "parameter",
    "response",
    "secured",
    "summary",
    "tag",
    "Array",
    "Binary",
    "Boolean",
    "Byte",
    "Component",
    "Date",
    "DateTime",
    "Double",
    "Email",
    "Float",
    "Integer",
    "Long",
    "Object",
    "Password",
    "String",
    "Time",
)


def _content_or_component(content):
    if isclass(content):
        spec = SpecificationBuilder()
        if spec._components["schemas"].get(content.__name__):
            content = definitions.Component(content)
    return content


@overload
def exclude(flag: bool = True, *, bp: Blueprint) -> None: ...


@overload
def exclude(flag: bool = True) -> Callable: ...


def exclude(flag: bool = True, *, bp: Optional[Blueprint] = None):
    if bp:
        for route in bp.routes:
            exclude(flag)(route.handler)
        return

    def inner(func):
        OperationStore()[func].exclude(flag)
        return func

    return inner


T = TypeVar("T")


def operation(name: str) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].name(name)
        return func

    return inner


def summary(text: str) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].describe(summary=text)
        return func

    return inner


def description(text: str) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].describe(description=text)
        return func

    return inner


def document(
    url: Union[str, definitions.ExternalDocumentation],
    description: Optional[str] = None,
) -> Callable[[T], T]:
    if isinstance(url, definitions.ExternalDocumentation):
        description = url.fields["description"]
        url = url.fields["url"]

    def inner(func):
        OperationStore()[func].document(url, description)
        return func

    return inner


def tag(*args: Union[str, definitions.Tag]) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].tag(*args)
        return func

    return inner


def deprecated(maybe_func=None) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].deprecate()
        return func

    return inner(maybe_func) if maybe_func else inner


def no_autodoc(maybe_func=None) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].disable_autodoc()
        return func

    return inner(maybe_func) if maybe_func else inner


def body(
    content: Any,
    *,
    validate: bool = False,
    body_argument: str = "body",
    **kwargs,
) -> Callable[[T], T]:
    body_content = _content_or_component(content)
    params = {**kwargs}
    validation_schema = None
    if isinstance(body_content, definitions.RequestBody):
        params = {**body_content.fields, **params}
        body_content = params.pop("content")

    if validate:
        if callable(validate):
            model = validate
        else:
            model = body_content
            validation_schema = generate_schema(body_content)

    def inner(func):
        @wraps(func)
        async def handler(*handler_args, **handler_kwargs):
            request = extract_request(*handler_args)

            if validate:
                try:
                    data = request.json
                    allow_multiple = False
                    allow_coerce = False
                except InvalidUsage:
                    data = request.form
                    allow_multiple = True
                    allow_coerce = True

                await do_validation(
                    model=model,
                    data=data,
                    schema=validation_schema,
                    request=request,
                    kwargs=handler_kwargs,
                    body_argument=body_argument,
                    allow_multiple=allow_multiple,
                    allow_coerce=allow_coerce,
                )

            retval = func(*handler_args, **handler_kwargs)
            if isawaitable(retval):
                retval = await retval
            return retval

        if func in OperationStore():
            OperationStore()[handler] = OperationStore().pop(func)
        OperationStore()[handler].body(body_content, **params)
        return handler

    return inner


@overload
def parameter(
    *,
    parameter: definitions.Parameter,
    **kwargs,
) -> Callable[[T], T]: ...


@overload
def parameter(
    name: None,
    schema: None,
    location: None,
    parameter: definitions.Parameter,
    **kwargs,
) -> Callable[[T], T]: ...


@overload
def parameter(
    name: str,
    schema: Optional[Union[type, Schema]] = None,
    location: Optional[str] = None,
    parameter: None = None,
    **kwargs,
) -> Callable[[T], T]: ...


def parameter(
    name: Optional[str] = None,
    schema: Optional[Union[type, Schema]] = None,
    location: Optional[str] = None,
    parameter: Optional[definitions.Parameter] = None,
    **kwargs,
) -> Callable[[T], T]:
    if parameter:
        if name or schema or location:
            raise SanicException(
                "When using a parameter object, you cannot pass "
                "other arguments."
            )
    if not schema:
        schema = str
    if not location:
        location = "query"

    def inner(func: Callable):
        if parameter:
            # Temporary solution convert in to location,
            # need to be changed later.
            fields = dict(parameter.fields)
            if "in" in fields:
                fields["location"] = fields.pop("in")
            OperationStore()[func].parameter(**fields)
        else:
            OperationStore()[func].parameter(name, schema, location, **kwargs)
        return func

    return inner


def response(
    status: Union[Literal["default"], int] = "default",
    content: Any = str,
    description: Optional[str] = None,
    *,
    response: Optional[definitions.Response] = None,
    **kwargs,
) -> Callable[[T], T]:
    if response:
        if (
            status != "default"
            or content is not str
            or description is not None
        ):
            raise SanicException(
                "When using a response object, you cannot pass "
                "other arguments."
            )

        status = response.fields["status"]
        content = response.fields["content"]
        description = response.fields["description"]

    def inner(func):
        OperationStore()[func].response(status, content, description, **kwargs)
        return func

    return inner


def secured(*args, **kwargs) -> Callable[[T], T]:
    def inner(func):
        OperationStore()[func].secured(*args, **kwargs)
        return func

    return inner


Model = TypeVar("Model")


def component(
    model: Optional[Model] = None,
    *,
    name: Optional[str] = None,
    field: str = "schemas",
) -> Callable[[T], T]:
    def wrap(m):
        return component(m, name=name, field=field)

    if not model:
        return wrap

    params = {}
    if name:
        params["name"] = name
    if field:
        params["field"] = field
    definitions.Component(model, **params)
    return model


def definition(
    *,
    exclude: Optional[bool] = None,
    operation: Optional[str] = None,
    summary: Optional[str] = None,
    description: Optional[str] = None,
    document: Optional[Union[str, definitions.ExternalDocumentation]] = None,
    tag: Optional[
        Union[
            Union[str, definitions.Tag], Sequence[Union[str, definitions.Tag]]
        ]
    ] = None,
    deprecated: bool = False,
    body: Optional[Union[dict[str, Any], definitions.RequestBody, Any]] = None,
    parameter: Optional[
        Union[
            Union[dict[str, Any], definitions.Parameter, str],
            list[Union[dict[str, Any], definitions.Parameter, str]],
        ]
    ] = None,
    response: Optional[
        Union[
            Union[dict[str, Any], definitions.Response, Any],
            list[Union[dict[str, Any], definitions.Response]],
        ]
    ] = None,
    secured: Optional[dict[str, Any]] = None,
    validate: bool = False,
    body_argument: str = "body",
) -> Callable[[T], T]:
    validation_schema = None
    body_content = None

    def inner(func):
        nonlocal validation_schema
        nonlocal body_content

        glbl = globals()

        if body:
            kwargs = {}
            content = body
            if isinstance(content, definitions.RequestBody):
                kwargs = content.fields
            elif isinstance(content, dict):
                if "content" in content:
                    kwargs = content
                else:
                    kwargs["content"] = content
            else:
                content = _content_or_component(content)
                kwargs["content"] = content

            if validate:
                kwargs["validate"] = validate
                kwargs["body_argument"] = body_argument

            func = glbl["body"](**kwargs)(func)

        if exclude is not None:
            func = glbl["exclude"](exclude)(func)

        if operation:
            func = glbl["operation"](operation)(func)

        if summary:
            func = glbl["summary"](summary)(func)

        if description:
            func = glbl["description"](description)(func)

        if document:
            kwargs = {}
            if isinstance(document, str):
                kwargs["url"] = document
            else:
                kwargs["url"] = document.fields["url"]
                kwargs["description"] = document.fields["description"]

            func = glbl["document"](**kwargs)(func)

        if tag:
            taglist = []
            op = (
                "extend"
                if isinstance(tag, (list, tuple, set, frozenset))
                else "append"
            )

            getattr(taglist, op)(tag)
            func = glbl["tag"](*taglist)(func)

        if deprecated:
            func = glbl["deprecated"]()(func)

        if parameter:
            paramlist = []
            op = (
                "extend"
                if isinstance(parameter, (list, tuple, set, frozenset))
                else "append"
            )
            getattr(paramlist, op)(parameter)

            for param in paramlist:
                kwargs = {}
                if isinstance(param, definitions.Parameter):
                    kwargs = param.fields
                    if "in" in kwargs:
                        kwargs["location"] = kwargs.pop("in")
                elif isinstance(param, dict) and "name" in param:
                    kwargs = param
                elif isinstance(param, str):
                    kwargs["name"] = param
                else:
                    raise SanicException(
                        "parameter must be a Parameter instance, a string, or "
                        "a dictionary containing at least 'name'."
                    )

                if "schema" not in kwargs:
                    kwargs["schema"] = str

                func = glbl["parameter"](**kwargs)(func)

        if response:
            resplist = []
            op = (
                "extend"
                if isinstance(response, (list, tuple, set, frozenset))
                else "append"
            )
            getattr(resplist, op)(response)

            if len(resplist) > 1 and any(
                not isinstance(item, definitions.Response)
                and not isinstance(item, dict)
                for item in resplist
            ):
                raise SanicException(
                    "Cannot use multiple bare custom models to define "
                    "multiple responses like openapi.definition(response=["
                    "MyModel1, MyModel2]). Instead, you should wrap them in a "
                    "dict or a Response object. See "
                    "https://sanic.dev/en/plugins/sanic-ext/openapi/decorators"
                    ".html#response for more details."
                )

            for resp in resplist:
                kwargs = {}
                if isinstance(resp, definitions.Response):
                    kwargs = resp.fields
                elif isinstance(resp, dict):
                    if "content" in resp:
                        kwargs = resp
                    else:
                        kwargs["content"] = resp
                else:
                    kwargs["content"] = resp

                if "status" not in kwargs:
                    kwargs["status"] = "default"

                func = glbl["response"](**kwargs)(func)

        if secured:
            func = glbl["secured"](secured)(func)

        return func

    return inner



from dataclasses import dataclass
from datetime import datetime

import attrs
import pytest

from msgspec import Struct
from pydantic import BaseModel
from pydantic.dataclasses import dataclass as pydataclass

from sanic_ext import openapi

from .utils import get_spec


@dataclass
class AlertDataclass:
    hit: dict[str, int]
    last_updated: datetime


@dataclass
class AlertResponseDataclass:
    alert: AlertDataclass
    rule_id: str


class AlertPydanticBaseModel(BaseModel):
    hit: dict[str, int]
    last_updated: datetime


class AlertResponsePydanticBaseModel(BaseModel):
    alert: AlertPydanticBaseModel
    rule_id: str


class AlertMsgspecBaseModel(Struct):
    hit: dict[str, int]
    last_updated: datetime


class AlertResponseMsgspecBaseModel(Struct):
    alert: AlertMsgspecBaseModel
    rule_id: str


@pydataclass
class AlertPydanticDataclass:
    hit: dict[str, int]
    last_updated: datetime


@pydataclass
class AlertResponsePydanticDataclass:
    alert: AlertPydanticDataclass
    rule_id: str


@attrs.define
class AlertAttrs:
    hit: dict[str, int]
    last_updated: datetime


@attrs.define
class AlertResponseAttrs:
    alert: AlertAttrs
    rule_id: str


@pytest.mark.parametrize(
    "AlertResponse,check_alert",
    (
        (AlertResponseDataclass, False),
        (AlertResponseAttrs, False),
        (AlertResponseMsgspecBaseModel, True),
        (AlertResponsePydanticBaseModel, True),
        (AlertResponsePydanticDataclass, True),
    ),
)
def test_pydantic_base_model(app, AlertResponse, check_alert):
    @app.get("/")
    @openapi.definition(
        body={"application/json": openapi.Component(AlertResponse)}
    )
    async def handler(_): ...

    spec = get_spec(app)
    alert_response_name = AlertResponse.__name__
    alert_name = "Alert" + alert_response_name[13:]

    assert spec["paths"]["/"]["get"]["requestBody"] == {
        "content": {
            "application/json": {
                "schema": {
                    "$ref": f"#/components/schemas/{alert_response_name}"
                }
            }
        }
    }
    assert alert_response_name in spec["components"]["schemas"]

    if check_alert:
        assert alert_name in spec["components"]["schemas"]
        assert spec["components"]["schemas"][alert_response_name][
            "properties"
        ]["alert"] == {"$ref": f"#/components/schemas/{alert_name}"}



from typing import Any, Callable

from sanic_ext.exceptions import ValidationError

from .check import check_data
from .clean import clean_data


try:
    from pydantic import ValidationError as PydanticValidationError

    VALIDATION_ERROR: tuple[type[Exception], ...] = (
        TypeError,
        PydanticValidationError,
    )
except ImportError:
    VALIDATION_ERROR = (TypeError,)


def validate_body(
    validator: Callable[[type[Any], dict[str, Any]], Any],
    model: type[Any],
    body: dict[str, Any],
) -> Any:
    try:
        return validator(model, body)
    except VALIDATION_ERROR as e:
        raise ValidationError(
            f"Invalid request body: {model.__name__}. Error: {e}",
            extra={"exception": str(e)},
        ) from None


def _msgspec_validate_instance(model, body, allow_coerce):
    import msgspec

    try:
        data = clean_data(model, body) if allow_coerce else body
        return msgspec.convert(data, model)
    except msgspec.ValidationError as e:
        # Convert msgspec.ValidationError into TypeError for consistent
        # behaviour with _validate_instance
        raise TypeError(str(e))


def _validate_instance(model, body, allow_coerce):
    data = clean_data(model, body) if allow_coerce else body
    return model(**data)


def _validate_annotations(model, body, schema, allow_multiple, allow_coerce):
    return check_data(model, body, schema, allow_multiple, allow_coerce)
